#!/usr/bin/python
#############################################################################
#
# XWisp
#
# see http://www.voti.nl/xwisp
#
# 19-NOV-2002 v 1.05
#    use of higher baudrates
#    PASS repaired
#    higher serial ports (COM9..) can now be used on Windows
#
# 19-OCT-2002 v 1.04
#    12F629 and 12F675 swap corrected
#    DUMP added to on-line help
#    CLOSE, CLEAR
#    WAIT corrected
#    RTS, DTR take immediate effect
#
# 30-SEP-2002 v 1.03
#    tested with jython, should work on posix
#
# 14-SEP-2002 v 1.02
#    16F877A added
#    target detection changed
#    MACRO bug corrected
#
# 05-SEP-2002 v 1.01
#    communication speedup
#
# 04-SEP-2002 v 1.00
#    first version
#
#############################################################################
#
# (c) 2002 Wouter van Ooijen / voti
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND ANY EXPRESS 
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
# THE POSSIBILITY OF SUCH DAMAGE.
#
#############################################################################

Version = '1.05'


#############################################################################
#
# misc
#
#############################################################################

def Split_Quoted( String ):
   List = []
   Item = ''
   Escape = 0
   Quote = ''
   for Char in String:
      #print Char, '"'.find( Char ), Quote, Item, List
      if Escape:
         Item = Item + Char
         Escape = 0
      # ecaping is a bit troublesome (Windows/Linux!)
      #elif Char == '\\':
      #   Escape = 1
      elif Char == Quote:
         #print 'Q]'
         List.append( Item )
         Item = ''
         Quote == ''      
      elif '"'.find( Char ) <> -1:
         #print 'Q['
         Quote = Char
      elif Quote <> '':
         #print 'Qq'
         Item = Item + Char
      elif " \t".find( Char )<> -1:
         if Item <> '':
            List.append( Item )
            Item = ''
      else:
         Item = Item + Char
   if Item <> '':
      List.append( Item )
   return List
   
# must be possible tro do this nicer...   
def Repeat( Char, N ):
   Result = ''
   for i in range( 0, N ):
      Result = Result + Char
   return Result   


#############################################################################
#
# intel hex file read, write, print, merge, etc.
#
#############################################################################

class Hex_Image:
   "read, write and manipulate Intel hex format files"

   class Hex_Image_Format_Error( IOError ):
      "errors raised by Hex_Image, derived from IOError"
      pass

   def Raise( self, Message ):
      "raise a Hex_Image_Format_Error with suitable prefix"
      raise self.Hex_Image_Format_Error, self.Error_Prefix + Message

   def __init__( self, File = None, Stride = 1, Suffixes = [ 'hex' ] ):
      "initialize, optionally specify stride and/or file to read from"
      self.Error_Prefix = ''
      self.Clear()
      self.Stride = Stride
      self.Suffixes = Suffixes
      if File <> None:
         self.Read( File )

   def Clear( self ):
      "clear all data, but do not modify stride"
      self.Data = {}
      self.Extended_Linear = 0

   def Set( self, Address, Value ):
      "set value at specified address"
      for Offset in range( 0, self.Stride ):
         self.Data[ Address * self.Stride + Offset ] = Value % 256
         Value = Value / 256;

   def Erase( self, Address ):
      for Offset in range( 0, self.Stride ):
         A = Address * self.Stride + Offset
         if self.Data.has_key( A ):
            del self.Data[ A ]

   def Get( self, Address, Default = None ):
      "return value at specified address"
      Value = 0
      for Offset in range( self.Stride - 1, -1, -1 ):
         Value = ( 256 * Value ) + \
            self.Data.get( self.Stride * Address + Offset, Default )
      return Value

   def Get_Hex( self, Address, Default = None ):
      "return hex (string) representation of value ar specified address"
      return \
         ( '%%0%dX' % ( 2 * self.Stride )) % \
         self.Get( Address, Default )

   def __Has_Range( self, Base, Count ):
      "check whether all addresses in Base .. Base + Count are present"
      for Address in range( Base, Base + Count ):
         if not self.Data.has_key( Address ):
            return 0
      return 1

   def Has( self, Address ):
      "check whether Address is present, taking stride into account"
      return self.__Has_Range( Address * self.Stride, self.Stride )

   def Addresses( self ):
      "return sorted list of all addresses"
      List = self.Data.keys()
      List.sort()
      return [ A / self.Stride for A in List \
         if (( A % self.Stride == 0 ) & \
            self.__Has_Range( A, self.Stride ))]

   def __Try_Open( self, File_Name, Suffixes ):
      if Suffixes == None:
         Suffixes = self.Suffixes
      try:
         File = open( File_Name, 'r' )
         return File
      except:
         pass
      for Suffix in Suffixes:
         try:
            File = open( File_Name + '.' + Suffix, 'r' )
            return File
         except:
            pass
      File = open( File_Name, 'r' )

   def Read( self, File_Name, Suffixes = None ):
      "read from a file"
      File = self.__Try_Open( File_Name, Suffixes )
      Save_Stride = self.Stride
      self.Stride = 1
      Linear_Base = 0
      try:
         Nr = 0
         for Line in File.readlines():
            Nr = Nr + 1
            self.Error_Prefix = File_Name + ' line ' + str( Nr ) + ' '
            while Line[-1:] < ' ':
               Line = Line[:-1]
            if Line[0] == ';':
                continue
            if Line[0] <> ':':
               self.Raise( 'does not start with a colon' )
            if len( Line ) % 2 <> 1:
               self.Raise( 'does not contain an odd number of characters' )
            Length = int( Line[1:3], 16 )
            Address = int( Line[3:7], 16 )
            Type = Line[7:9]
            Checksum = ( Length + Address / 256 + \
               Address % 256 + int( Type )) % 256
            Address = Address + Linear_Base
            Line = Line[9:]
            if Type == '00':
               # data
               while len(Line) > 2:
                  Data = int( Line[0:2], 16 )
                  self.Set( Address, Data )
                  Line = Line[2:]
                  Address = Address + 1
                  Checksum = ( Checksum + Data ) % 256

            elif Type == '01':
               # end of file
               pass

            elif Type == '02':
               # extended segment
                self.Raise( 'Type=2' )

            elif Type == '03':
               # start segment
               self.Raise( 'Type=3' )

            elif Type == '04':
               # extended linear
               Value = int( Line[0:4], 16 )
               Linear_Base = 0x10000 * Value
               Checksum = ( Checksum + Value / 256 ) + ( Value % 256 )
               self.Extended_Linear = 1
               Line = Line[4:]

            elif Type == '05':
               # start linear
               self.Raise( 'Type=5' )

            else:
               self.Raise( 'unknown type=' + Type )

            if ( Checksum + int( Line, 16 )) % 256 <> 0:
               self.Raise(
                  'checksum found=' + Line +
                  ' expected=' + ( '%02X' % (( 256 - Checksum ) % 256 )))

      finally:
         self.Stride = Save_Stride
         File.close()

   def Add( self, Image ):
      "add content of another image"
      for Address in Image.Data.keys:
         self.Data[ Address ] = Image.Data.get( Address )

   def __Hex_Line( self, Count, Base, Type, Data, Checksum ):
      "create data line for writing"
      Checksum = Checksum + Count + (( Base / 256 ) % 256 ) + ( Base % 256 )
      Checksum = ( 256 - ( Checksum % 256 )) % 256
      return ':' + \
         ( '%02X' % Count ) + \
         ( '%04X' % ( Base % 0x10000 )) + \
         ( '%02X' % Type ) + \
         Data + \
         ( '%02X' % Checksum )

   def Write( self, File_Name, Extended_Linear = None ):
      "write to a hex file"
      import os
      File_Name = os.path.normpath( File_Name )
      try:
         os.makedirs( os.path.dirname( File_Name ))
      except:
         pass
      File = open( File_Name, 'w' )
      Save_Stride = self.Stride
      self.Stride = 1
      try:
         if Extended_Linear == None:
            Extended_Linear = self.Extended_Linear
         if Extended_Linear:
            File.write( ':020000040000FA\n' )
         Count = 10000
         Line = ''
         Last = 0
         for Address in self.Addresses():
            if ( Count >= 16 ) \
            or ( Address <> Last + 1 ) \
            or ( Address % 16 == 0) \
            or ( Address / 0x10000 ) <> ( Last / 0x10000 ):
               if Line <> '':
                  File.write( self.__Hex_Line(
                     Count, Base, 0x00, Line, Checksum ) + '\n')
               Base = Address
               Count = 0
               Line = ''
               Checksum = 0
            if Address / 0x10000 <> Last / 0x10000:
               Segment = Address / 0x10000
               File.write( self.__Hex_Line(
                  0x02, 0x0000, 0x04,
                  ( '%04X' % Segment ),
                  2 + 4 + Segment / 256 + Segment % 256 ) + '\n')
            Line = Line + self.Get_Hex( Address )
            Checksum = ( Checksum + self.Get( Address )) % 256
            Count = Count + 1
            Last = Address
         if Line <> '':
            File.write( self.__Hex_Line(
               Count, Base, 0x00, Line, Checksum ) + '\n')
         File.write( ':00000001FF\n' )
      finally:
         self.Stride = Save_Stride
         File.close()

   def __str__( self ):
      "return printeable representation"
      Last = -10
      Count = 0
      Result = ''
      for Address in self.Addresses():
         if ( Address <> Last + 1 ) | ( Count == ( 16 / self.Stride )):
            if Result <> '':
               Result = Result + '\n'
            Result = Result + ( '%04X:' % Address )
            Count = 0
         Result = Result + ' ' + self.Get_Hex( Address )
         Count = Count + 1
         Last = Address
      if Result == '':
         Result = 'image is empty'
      return ( 'stride=%d\n' % self.Stride ) + Result

   def String( self ):
      return self.__str__()

   def Clone( self ):
      "return cloned copy"
      import copy
      return copy.deepcopy( self )

   def Compare(
      self,
      Image,
      Name1 = 'image1',
      Name2 = 'image2',
      Range = None
   ):
      "return None or string describing first difference"
      List_1 = self.Addresses()
      List_2 = Image.Addresses()
      if self.Stride != Image.Stride:
         return "strides are different: $s=%d %s=%d" % \
            Name1, Name2, self.Stride, Image.Stride
      for Address in List_1:
         if ( Range == None ) or Address in Range:
            if len( List_2 ) == 0:
               return "address %04X in %s but not in %s" % \
                  ( Address, Name1, Name2 )
            Other = List_2.pop( 0 )
            if Address < Other:
               return "address %04X in %s but not in %s" % \
                  ( Address, Name1, Name2 )
            if Address > Other:
               return "address %04X in %s but not in %s" % \
                  ( Other, Name2, Name1 )
            if self.Get( Address ) != Image.Get( Other ):
               return "different data at address %04X: %s=%s %s=%s" % \
                  ( Address, \
                  Name1, self.Get_Hex( Address ), \
                  Name2, Image.Get_Hex( Address ) )
      return None


#############################################################################
#
# WBus access
#
#############################################################################

class WBus:
   "access to a WBus serial 'bus'"

   class WBus_Protocol_Error( IOError ):
      "protocol errors raised by WBus, derived from IOError"
      pass

   class WBus_Command_Failed( IOError ):
      "command failure error raised by WBus, derived from IOError"
      pass

   class WBus_Use_Error( IOError ):
      "use errors raised by WBus, derived from IOError"
      pass

   def Log( self, String ):
      "debug log"
      import time, sys
      self.Console.Print( str( "%04.3f" % time.clock() ) + ' ' + String )

   def Set_RTS( self, Value ):
      "set RTS line"
      if self.Debug:
         if Value <> None:
            self.Log( 'self.SetRTS( %d )' % Value )
      self.RTS = Value
      if self.RTS != None:
         self.Port.setRTS( self.RTS )

   def Set_DTR( self, Value ):
      "set DTR line"
      if self.Debug:
         if Value <> None:
            self.Log( 'self.SetDTR( %d )' % Value )
      self.DTR = Value
      if self.DTR != None:
        self.Port.setDTR( self.DTR )

   def __init__(
      self,
      Console,
      Port = None,
      ID = 0,
      Baudrate = 19200,
      DTR = None,
      RTS = None,
      Debug = 0
   ):
      '''initialize and connect Port.
      Port, ID, Baudrate etc. can be specified'''
      import serial, time

      self.Console = Console
      self.Com_Port = Port
      self.ID = ID
      self.Baudrate = Baudrate
      self.DTR = DTR
      self.RTS = RTS
      self.Debug = 0
      self.Error_Prefix = ''
      self.Fast = 0

      # connect the serial port
      if self.Debug:
         self.Log(
            'WBus( Port=%s Baudrate=%s )'
            % ( str( self.Com_Port ), str( self.Baudrate )))
      self.Port = serial.Serial(
         port      = self.Com_Port,
         baudrate  = self.Baudrate,
         stopbits  = serial.STOPBITS_ONE,
         timeout   = 2.0 )

      # set handshake lines, perform reset only if no DTR value set
      self.Set_RTS( RTS )
      if self.DTR <> None:
         self.Port.setDTR( self.DTR )
      else:
         self.Set_DTR( 1 )
         time.sleep( 0.005 )
         self.Set_DTR( 0 )

   def Connect(
      self,
      ID = None,
      DTR = None,
      RTS = None
   ):
      '''connect to WBus device.
      ID, Baudrate, DTR, RTS can be specified'''

      # handle parameters
      if ID <> None:
         self.ID = ID
      self.Set_RTS( RTS )
      self.Set_DTR( DTR )

      # send break to get devices to attention or active state
      self.Port.sendBreak( duration = 0.200 )

      # test for echoing hardware
      self.Clear()
      self.Port.write( chr( 255 ))
      Echo = self.Port.read( 1 )
      self.Echoing = ( Echo == chr( 255 ))

      # activate device
      self.Send_Slowly(( '%04X' % self.ID ) + 'h' )

      # get device type and version
      self.Clear()
      self.Send_Expect( 't' )
      self.Type = self.Get()
      self.Send_Expect( 'v' )
      self.Version = self.Get()

   def Close( self ):
      "close connection and serial port"
      if self.Debug:
         self.Log( 'self.Close()' )
      self.Port.close()

   def Clear( self ):
      "clear input buffer"
      import time
      if self.Debug:
         self.Log( 'Clear()' )
      time.sleep( 0.1 )
      self.Port.flushInput()
      self.Port.flushOutput()

   def Send_Char( self, Char ):
      "send a char, observe timing, set high bit when hardware echoes"
      import time
      #time.sleep( 0.002 )
      if self.Debug:
         self.Log( "Send_Char( '%s' )" % Char )
      if self.Echoing and ( Char != '' ) :
         Char = chr( ord( Char ) + 128 )
      # print 'x1'
      self.Port.write( Char )
      # print 'x2'
      if self.Debug:
         self.Log( 'Send_Char done' )

   def Receive_Char( self ):
      "receive a char, skip hardware echo"
      if self.Debug:
         self.Log( 'Receive_Char()' )
      Result = self.Port.read( 1 )
      if self.Echoing:
         Result = self.Port.read( 1 )
      if self.Debug:
         self.Log( "Received = '%s'" % Result )
      return Result

   def Send_Slowly( self, String ):
      "send slowly"
      import time
      if self.Debug:
         self.Log( 'Send_Slowly( "' + String + '" )' )
      for Char in String:
         time.sleep( 0.05 )
         self.Send_Char( Char )

   def Send_Receive( self, Char ):
      "send a char and expect a response from the device"
      if self.Debug:
         self.Log( 'Send_Receive( "' + Char + '" )' )
      self.Send_Char( Char )
      Response = self.Receive_Char()
      if Response == '?':
         raise self.WBus_Command_Failed, \
            "send=%c" % Char
      if self.Debug:
         self.Log( "Response='%s'" % Response )
      return Response

   def Send_Expect( self, String ):
      "send and expect echo from the device"
      import time
      if self.Debug:
         self.Log( 'Send_Expect( "' + String + '" )' )
      if self.Fast: 
         self.Port.write( String )      
         Reply = self.Port.read( len( String ))
         if Reply != String.upper():
            raise self.WBus_Protocol_Error, \
               "send='" + String + "' received='" + Reply + "'"
      else:
         for Char in String:
            Reply = self.Send_Receive( Char )
            if Reply != Char.upper():
               raise self.WBus_Protocol_Error, \
                  "send='" + Char + "' received='" + Reply + "'"

   def Get( self, N = None ):
      "get string or fixed size response from the device"
      if self.Debug: self.Log( 'Get( N=' + str( N ) + ' )' )
      Result = ''
      if N != None:
         if self.Fast:
            self.Port.write( Repeat( 'n', N ))
            Result = self.Port.read( N )
         else:
            for Dummy in range( N ):
               Response = self.Send_Receive( 'n' )
               Result = Result + Response
      else:
         Response = self.Send_Receive( 'n' )
         if ( Response != ' ' ):
            for Dummy in [ 1, 2, 3 ]:
               Response = Response + self.Send_Receive( 'n' )
            return Response
         else:
            Response = ''
            while ( Response != ' ' ) and ( len( Response ) < 32 ):
               Result = Result + Response
               Response = self.Send_Receive( 'n' )
      return Result


#############################################################################
#
# Target information
#
#############################################################################

Algorithm_PIC16   = 0
Algorithm_PIC16A  = 1
Algorithm_PIC16B  = 2
Algorithm_PIC18   = 3
Algorithm_PIC16C  = 4

Region_Code   = 10
Region_Data   = 11
Region_ID     = 12
Region_Fuses  = 13

Region_Name = {
   Region_Code  : "Code",
   Region_Data  : "Data",
   Region_ID    : "ID",
   Region_Fuses : "Fuses" }

# fuses must be last for Write_Verify
All_Regions = [ Region_Code, Region_Data, Region_ID, Region_Fuses ]
Programming_Regions = [ Region_Code, Region_Data, Region_Fuses ]

_PIC_Types_By_ID    = {}
_PIC_Types_By_Name  = {}

class PIC_Range:
   "describes a PIC store region"

   def __init__( self, Base, Start = None, End = None, Ignore = [] ):
     if Start == None:
         self.Base = 0
         self.Start = 0
         self.End = Base - 1
         self.Ignore = [ 0x3FFF ]
     else:
         self.Base = Base
         self.Start = Start
         self.End = End
         self.Ignore = Ignore

   def Range( self, Reverse = 0 ):
      Result = range( self.Start, self.End + 1 )
      if Reverse:
         Result.reverse()
      return Result

class PIC_EEPROM( PIC_Range ):
   "describes PIC EEPROM data storage"
   def __init__( self, Size ):
      PIC_Range.__init__( self, 0x2100, 0x2100, 0x2100 + Size - 1, [ 0x00FF ] )

class PIC_Code( PIC_Range ):
   "describes PIC Flash code storage"
   def __init__( self, Size ):
      PIC_Range.__init__( self, Size )

class PIC_One_Type:
   "describes a single PICmicro type"

   def __init__(
      self,
      Name = None,
      Shorthands = None,
      Algorithm = None,
      Code = None, Data = None, ID = None, Fuses = None,
      Protect = None,
      ID_Start = None, ID_Value = None,
      Speed = None
   ):
      self.Name = Name
      self.Shorthands = Shorthands
      _PIC_Types_By_Name[ Name ] = self
      for Short in Shorthands:
         _PIC_Types_By_Name[ Short ] = self
      _PIC_Types_By_ID[ ID_Value ] = self
      self.Algorithm = Algorithm
      self.Speed = Speed
      if ( Algorithm == Algorithm_PIC16 ) or \
      ( Algorithm == Algorithm_PIC16A ) or \
      ( Algorithm == Algorithm_PIC16B ) or \
      ( Algorithm == Algorithm_PIC16C ) :
         ID        = PIC_Range( 0x2000, 0x2006, 0x2006 )
         Fuses     = PIC_Range( 0x2000, 0x2007, 0x2007 )
      elif ( Algorithm == Algorithm_PIC18 ):
         ID        = PIC_Range( 0x000000, 0x3FFFFE, 0x3FFFFF )
         Fuses     = PIC_Range( 0x000000, 0x300000, 0x30000D )
      else:
         raise IOError, 'internal error: invalid algorithm'
      if Data == None:
         Data = PIC_Range( 0, 0, -1 )
      self.Region = {}
      self.Region[ Region_Code ]   = Code
      self.Region[ Region_Data ]   = Data
      self.Region[ Region_ID ]     = ID
      self.Region[ Region_Fuses ]  = Fuses
      self.Protect = Protect
      self.ID_Start = 0x20
      self.Fuses_Address = 0x2007
      if Algorithm == Algorithm_PIC18:
         self.Factor = 0x100
         self.Stride = 1
         self.Write_Cluster = 8
      else:
         self.Factor = 0x10000
         self.Stride = 2
         self.Write_Cluser = 1

   def In_Region( self, Region, Address ):
      # print Address, self.Region[ Region].Start, self.Region[ Region].End
      return ( Address >= self.Region[ Region].Start ) and \
         ( Address <= self.Region[ Region].End )

def PIC_Type_From_Name( Name, Default = None ):
   return _PIC_Types_By_Name.get( Name.lower(), Default )

def PIC_Type_From_ID( ID, Default = None ):
   return _PIC_Types_By_ID.get( ID, Default )

PIC_One_Type(
   Name        = '12f629',
   Shorthands  = [ 'f629', '629' ],
   Algorithm   = Algorithm_PIC16C,
   Code        = PIC_Range( 1024 ),
   Data        = PIC_EEPROM( 128 ),
   Protect     = 0x000F,
   ID_Value    = 0x0F80
   )
   
PIC_One_Type(
   Name        = '12f675',
   Shorthands  = [ 'f675', '675' ],
   Algorithm   = Algorithm_PIC16C,
   Code        = PIC_Range( 1024 ),
   Data        = PIC_EEPROM( 128 ),
   Protect     = 0x000F,
   ID_Value    = 0x0FC0
   )

PIC_One_Type(
   Name        = '16c84',
   Shorthands  = [ 'c84' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Range( 1024 ),
   Data        = PIC_EEPROM( 64 ),
   Protect     = 0x000F,
   ID_Value    = -1
   )

PIC_One_Type(
   Name        = '16f83',
   Shorthands  = [ 'f83', '83' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 512 ),
   Data        = PIC_EEPROM( 64 ),
   Protect     = 0x3FF0,
   ID_Value    = -1
   )

PIC_One_Type(
   Name        = '16f84',
   Shorthands  = [ 'f84', 'x84' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 1024 ),
   Data        = PIC_EEPROM( 64 ),
   Protect     = 0x3FF0,
   ID_Value    = -1
   )

PIC_One_Type(
   Name        = '16f84a',
   Shorthands  = [ 'f84a', '84a' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 1024 ),
   Data        = PIC_EEPROM( 64 ),
   Protect     = 0x3FF0,
   ID_Value    = 0x0560
   )

PIC_One_Type(
   Name        = '16f627',
   Shorthands  = [ 'f627', '627' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 1024 ),
   Data        = PIC_EEPROM( 224 ),
   Protect     = 0x3D00,
   ID_Value    = 0x07A0
   )

PIC_One_Type(
   Name        = '16f628',
   Shorthands  = [ 'f628', '628' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 2048 ),
   Data        = PIC_EEPROM( 224 ),
   Protect     = 0x3D00,
   ID_Value    = 0x07C0,
   Speed       = 80
   )

PIC_One_Type(
   Name        = '16f73',
   Shorthands  = [ 'f73', '73' ],
   Algorithm   = Algorithm_PIC16A,
   Code        = PIC_Code( 4096 ),
   Protect     = 0x0010,
   ID_Value    = 0x0600
   )

PIC_One_Type(
   Name        = '16f74',
   Shorthands  = [ 'f74', '74' ],
   Algorithm   = Algorithm_PIC16A,
   Code        = PIC_Code( 4096 ),
   Protect     = 0x0010,
   ID_Value    = 0x0620
   )

PIC_One_Type(
   Name        = '16f76',
   Shorthands  = [ 'f76', '76' ],
   Algorithm   = Algorithm_PIC16A,
   Code        = PIC_Code( 8192 ),
   Protect     = 0x0010,
   ID_Value    = 0x0640
   )

PIC_One_Type(
   Name        = '16f77',
   Shorthands  = [ 'f77', '77' ],
   Algorithm   = Algorithm_PIC16A,
   Code        = PIC_Code( 8192 ),
   Protect     = 0x0010,
   ID_Value    = 0x0660
   )

PIC_One_Type(
   Name        = '16f870',
   Shorthands  = [ 'f870', '870' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 2048 ),
   Data        = PIC_EEPROM( 64 ),
   Protect     = 0x3130,
   ID_Value    = 0x0D00
   )

PIC_One_Type(
   Name        = '16f871',
   Shorthands  = [ 'f871', '871' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 2048 ),
   Data        = PIC_EEPROM( 64 ),
   Protect     = 0x3130,
   ID_Value    = 0x0D20
   )

PIC_One_Type(
   Name        = '16f872',
   Shorthands  = [ 'f872', '872' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 2048 ),
   Data        = PIC_EEPROM( 64 ),
   Protect     = 0x3130,
   ID_Value    = 0x08E0
   )

PIC_One_Type(
   Name        = '16f873',
   Shorthands  = [ 'f873', '873' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 4096 ),
   Data        = PIC_EEPROM( 128 ),
   Protect     = 0x3130,
   ID_Value    = 0x0960
   )

PIC_One_Type(
   Name        = '16f873A',
   Shorthands  = [ 'f873a', '873a' ],
   Algorithm   = Algorithm_PIC16B,
   Code        = PIC_Code( 4096 ),
   Data        = PIC_EEPROM( 128 ),
   Protect     = 0x3130,
   ID_Value    = 0x0E40
   )

PIC_One_Type(
   Name        = '16f874',
   Shorthands  = [ 'f874', '874' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 4096 ),
   Data        = PIC_EEPROM( 128 ),
   Protect     = 0x3130,
   ID_Value    = 0x0920
   )

PIC_One_Type(
   Name        = '16f874A',
   Shorthands  = [ 'f874a', '874a' ],
   Algorithm   = Algorithm_PIC16B,
   Code        = PIC_Code( 4096 ),
   Data        = PIC_EEPROM( 128 ),
   Protect     = 0x0010,
   ID_Value    = 0x0E60
   )

PIC_One_Type(
   Name        = '16f876',
   Shorthands  = [ 'f876', '876' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 8192 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x09E0
   )

PIC_One_Type(
   Name        = '16f876A',
   Shorthands  = [ 'f876a', '876a' ],
   Algorithm   = Algorithm_PIC16B,
   Code        = PIC_Code( 8192 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0E00
   )

PIC_One_Type(
   Name        = '16f877',
   Shorthands  = [ 'f877', '877' ],
   Algorithm   = Algorithm_PIC16,
   Code        = PIC_Code( 8192 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x09A0
   )

PIC_One_Type(
   Name        = '16f877a',
   Shorthands  = [ 'f877a', '877a' ],
   Algorithm   = Algorithm_PIC16B,
   Code        = PIC_Code( 8192 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0E20
   )

PIC_One_Type(
   Name        = '18f242',
   Shorthands  = [ 'f242', '242' ],
   Algorithm   = Algorithm_PIC18,
   Code        = PIC_Code( 16 * 1024 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0480
   )

PIC_One_Type(
   Name        = '18f248',
   Shorthands  = [ 'f248', '248' ],
   Algorithm   = Algorithm_PIC18,
   Code        = PIC_Code( 16 * 1024 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0800
   )

PIC_One_Type(
   Name        = '18f252',
   Shorthands  = [ 'f252', '252' ],
   Algorithm   = Algorithm_PIC18,
   Code        = PIC_Code( 32 * 1024 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0400
   )

PIC_One_Type(
   Name        = '18f258',
   Shorthands  = [ 'f258', '258' ],
   Algorithm   = Algorithm_PIC18,
   Code        = PIC_Code( 32 * 1024 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0840
   )

PIC_One_Type(
   Name        = '18f442',
   Shorthands  = [ 'f442', '442' ],
   Algorithm   = Algorithm_PIC18,
   Code        = PIC_Code( 16 * 1024 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x04A0
   )

PIC_One_Type(
   Name        = '18f448',
   Shorthands  = [ 'f448', '448' ],
   Algorithm   = Algorithm_PIC18,
   Code        = PIC_Code( 16 * 1024 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0820
   )

PIC_One_Type(
   Name        = '18f452',
   Shorthands  = [ 'f452', '452' ],
   Algorithm   = Algorithm_PIC18,
   Code        = PIC_Code( 32 * 1024 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0420
   )

PIC_One_Type(
   Name        = '18f458',
   Shorthands  = [ 'f458', '458' ],
   Algorithm   = Algorithm_PIC18,
   Code        = PIC_Code( 32 * 1024 ),
   Data        = PIC_EEPROM( 256 ),
   Protect     = 0x3130,
   ID_Value    = 0x0860
   )


#############################################################################
#
# Console abstraction
#
#############################################################################

class Dos_Console:
   def __init__( self ):
      self.Erase = 0

   def Beep( self ):
      try:
         import winsound
         winsound.Beep( 440, 400 )
      except:
         self.Print( "sorry, could not beep" )

   def Print( self, String, Progress = 0 ):
      import sys
      if self.Erase:
         sys.stdout.write( "\r%60s\r" % '' )
         self.Erase = 0
      self.Erase = Progress
      sys.stdout.write( String )
      if not Progress:
         sys.stdout.write( '\n' )
      sys.stdout.flush()
      
      # wxPython process events and update statusbar
      global app
      if app<>None:
         while app.Pending():
            app.Dispatch()
         app.win.UpdateStatusbar(String)

   def Print_String( self, String ):
      import sys
      sys.stdout.write( String )
      sys.stdout.flush()

   def Set_Raw( self ):
      import os
      if os.name == 'nt':
         pass
      elif os.name == 'posix':
         # Untested code derived from the Python FAQ....
         import termios, TERMIOS, sys
         self.fd = sys.stdin.fileno()
         self.Old_Attr = termios.tcgetattr( self.fd )
         new = termios.tcgetattr( self.fd )
         new[3] = new[3] & ~TERMIOS.ICANON & ~TERMIOS.ECHO
         new[6][TERMIOS.VMIN] = 1
         new[6][TERMIOS.VTIME] = 0
         termios.tcsetattr( self.fd, TERMIOS.TCSANOW, new )
      else:
         raise IOError, "TTY not implemented for %s" % sys.platform

   def Set_Cooked( self ):
      import os
      if os.name == 'posix':
         import termios, TERMIOS, sys
         termios.tcsetattr( self.fd, TERMIOS.TCSAFLUSH, self.Old_Attr )

   def Get_Key( self ):
      import os
      if os.name == 'nt':
         import msvcrt
         if not msvcrt.kbhit():
            return ''
         while 1:
            z = msvcrt.getch()
            if z == '\0' or z == '\xe0':
                # was a functions key prefix
                msvcrt.getch()
            else:
                return z
      elif os.name == 'posix':
         return os.read( self.fd, 1 )

   def Title( self, Title ):
      pass

#############################################################################
#
# Programming over a WBus
#
#############################################################################


class Bus_Target:
   "abstracts a programmable target"

   def __init__(
      self,
      Console,
      Bus = None,
      Type = None,
      DTR = None,
      RTS = None,
   ):
      "initialize. Bus and (target) Type can be specified"
      self.Bus = Bus
      self.Type = Type
      self.Region = None
      self.Address = 0
      self.New_Region = 0
      self.Debug = 0
      self.DTR = DTR
      self.RTS = RTS
      self.Console = Console
      self.Connected = 0

   def Print( self, String, Progress = 0 ):
      self.Console.Print( String, Progress )

   def Log( self, String ):
      "debug log"
      import time, sys
      if self.Debug:
         self.Print( str( "%04.3f" % time.clock() ) + ' ' + String )

   def Connect( self, Bus = None, Type = None, DTR = None, RTS = None ):
      "connect to the programmer and target."

      # handle arguments"
      if Bus <> None:
         self.Bus = Bus
      if Type <> None:
         self.Type = Type
      if DTR <> None:
         self.DTR = DTR
      if RTS <> None:
         self.RTS = RTS

      # for debugging
      self.Log( "Connect( ... )" )

      # connect
      self.Bus.Connect( DTR = self.DTR, RTS = self.RTS )

      # check whether the device is supported
      if self.Bus.Type == 'WISP':
         pass
      elif self.Bus.Type == 'Wisp628':
         pass
      elif self.Bus.Type == 'WLdr':
         pass
      else:
         raise IOError, 'device "' + self.Bus.Type + '" is not supported'
         
      self.Connected = 1

   def Close( self ):
      "close WBus connection"
      self.Log( "Close()" )
      self.Bus.Close()
      self.Connected = 0

   def __Start_Programming( self ):
      "start programming at the start of self.Region"
      import time
      self.Log( "__Start_Programming() " + \
         "Region=" + Region_Name[ self.Region ])
      self.Address = self.Type.Region[ self.Region ].Start
      Cmd = ''
      if self.Type.Algorithm == Algorithm_PIC16:
         if self.Region == Region_Code:   Cmd = '000cx'
         if self.Region == Region_Data:   Cmd = '000dx'
         if self.Region == Region_ID:     Cmd = '000fx'
         if self.Region == Region_Fuses:  Cmd = '000fx'
      if self.Type.Algorithm == Algorithm_PIC16A:
         if self.Region == Region_Code:   Cmd = '001cx'
         if self.Region == Region_Data:   Cmd = '001dx'
         if self.Region == Region_ID:     Cmd = '001fx'
         if self.Region == Region_Fuses:  Cmd = '001fx'
      if self.Type.Algorithm == Algorithm_PIC16B:
         if self.Region == Region_Code:   Cmd = '002cx'
         if self.Region == Region_Data:   Cmd = '002dx'
         if self.Region == Region_ID:     Cmd = '002fx'
         if self.Region == Region_Fuses:  Cmd = '002fx'
      if self.Type.Algorithm == Algorithm_PIC16C:
         if self.Region == Region_Code:   Cmd = '004cx'
         if self.Region == Region_Data:   Cmd = '000dx'
         if self.Region == Region_ID:     Cmd = '004fx'
         if self.Region == Region_Fuses:  Cmd = '004fx'
      if self.Type.Algorithm == Algorithm_PIC18:
         if self.Region == Region_Code:   Cmd = '003cx'
         if self.Region == Region_Data:   Cmd = '003dx'
         if self.Region == Region_ID:     Cmd = '003fx'
         if self.Region == Region_Fuses:  Cmd = '003fx'
      if Cmd == '':
         raise IOError, \
            'internal error: unknown Algorithm or Region: ' + \
            str( self.Type.Algorithm ) + ' ' + str( self.Region )
      #time.sleep( 0.2 ) #why?
      self.Bus.Send_Expect( Cmd )
      self.Address = self.Type.Region[ self.Region ].Base

   def Set_Region( self, Region ):
      "switch interest to indicated region"
      self.Log( "Set_Region( %d )" % Region )
      self.Region = Region
      self.Address = -1
      self.New_Region = 1

   def __GoTo( self, Address ):
      "prepare read or write at self.Region and indicated address"
      self.Log(( "__GoTo( %04X )" % Address ) + \
         ( " self.Address=%04X" % self.Address ) + \
         ( " Region=" +  Region_Name[ self.Region ] ))
      if ( self.Address == -1 ) or ( Address < self.Address ):
         self.__Start_Programming()
      # the self.Region == condition should not be needed yet it is...
      if ( self.Type.Algorithm == Algorithm_PIC18 ) or \
         (( self.Bus.Type == 'Wisp628' ) and \
          ( float( self.Bus.Version ) > 1.02 ) and \
          ( self.Region == Region_Code ) and \
          (( Address - self.Address ) > 4 )):
         if Address <> self.Address:
            self.Bus.Send_Expect( '%06Xm' % Address )
            self.Address = Address
      else:
         while Address > self.Address:
            self.Bus.Send_Expect( 'i' )
            self.Address = self.Address + 1

   def Put( self, Address, Data ):
      "put (write) the indicated data at the indicated address"
      self.Log( "Put( %04X )" % Address )
      self.__GoTo( Address )
      self.Bus.Send_Expect(( '%04X' % Data ) + 'w' )

   def Get( self, Address ):
      "get (read) from the indicated address"
      self.Log( "Get( %04X )" % Address )
      self.__GoTo( Address )
      self.Bus.Send_Expect( 'r' )
      if self.Type.Algorithm == Algorithm_PIC18:
         self.Address = self.Address + 1
      return int( self.Bus.Get( 2 * self.Type.Stride ), 16)

   def Get_ID( self ):
      "get the target's device ID word"
      self.Log( "Get_ID()" )
      self.Set_Region( Region_ID )
      Result = 0
      for Address in self.Type.Region[ Region_ID ].Range( Reverse = 1):
         Result = Result * self.Type.Factor + self.Get( Address )
      self.Device_Code = Result
      self.Device_Revision = Result % self.Type.ID_Start
      self.Device_ID = Result - self.Device_Revision
      self.Log(
         "CODE=%04X REV=%04X ID=%04X" %
         ( self.Device_Code, self.Device_Revision, self.Device_ID ))

   def Identify( self, Expect = None ):
      "identify the connected PIC"
      self.Log( "Identify()" )
      if Expect == None:
         self.Type = PIC_Type_From_Name( '16c84' )
         self.Get_ID()
         self.Type = PIC_Type_From_ID( self.Device_ID, None )
         if self.Type == None:
            self.Device_Code_16 = self.Device_Code
            self.Type = PIC_Type_From_Name( '18f452' )
            self.Get_ID()
            self.Type = PIC_Type_From_ID( self.Device_ID, None )
      else:
         self.Type = PIC_Type_From_Name( Expect )
         self.Get_ID()
         self.Type = PIC_Type_From_ID( self.Device_ID, None )

   def Run( self, Cmd = '0000' ):
      "switch target to 'run', use indicated mode"
      self.Log( "Run( Cmd=%s )" % Cmd )
      self.Bus.Send_Expect( Cmd + 'g' )

   def Progress( self, Address ):
      "called for each address where actual read or write takes place"
      "default: log first address in each reagion, and every multiple of 16"
      import sys
      if self.New_Region or ( Address % 16 == 0 ) or ( self.Debug ):
         self.New_Region = 0
         Line = self.Operation + ' ' + \
            Region_Name[ self.Region ] + ' ' + \
            ( '%04X' % Address )
         self.Print( Line, Progress = 1 ) 

   def Write( self, Image, Regions = Programming_Regions ):
      "write the indicated regions of the indicated image to the target"
      self.Log( "Write( ... )" )
      self.Operation = 'write'
      for Region in Regions:
         self.Set_Region( Region )
         if ( self.Type.Algorithm == Algorithm_PIC18 ) \
         and ( Region == Region_Code ):
            Message = ''
            N = 0
            for Address in self.Type.Region[ Region ].Range():
               Message = Message + ( Image.Get_Hex( Address, 0xFF ))
               if Image.Has( Address ):
                  self.Progress( Address )
                  N = N + 1
               if ( Address + 1 ) % self.Type.Write_Cluster == 0:
                  if N > 0:
                     self.__GoTo( Address - self.Type.Write_Cluster + 1 )
                     self.Bus.Send_Expect( Message + 'w' )
                     self.Address = Address + 1
                  N = 0
                  Message = ''
         else:
            for Address in Image.Addresses():
               #print 'a=', Address
               if self.Type.In_Region( self.Region, Address ):
                  self.Progress( Address )
                  self.Put( Address, Image.Get( Address ))

   def Read( self, Regions = Programming_Regions, List = None, Ignore = 1 ):
      "read the indicated regions from the target"
      self.Log( "Read( ... )" )
      self.Operation = 'read'
      Image = Hex_Image( Stride = self.Type.Stride )
      for Region in Regions:
         self.Set_Region( Region )
         for Address in self.Type.Region[ Region ].Range():
            if ( List == None ) or ( List.Has( Address )):
               self.Progress( Address )
               Data = self.Get( Address )
               self.Log( "Read: %04X = %04X" % ( Address, Data ) )
               if ( not Ignore ) or \
               ( not Data in self.Type.Region[ Region ].Ignore ):
                  Image.Set( Address, Data )
      return Image

   def Verify( self, Image, Regions = Programming_Regions ):
      for Region in Regions:
         Readback = self.Read( [ Region ], List = Image, Ignore = 0 )
         Diff = Image.Compare(
            Readback, 'file', 'target',
            Range = self.Type.Region[ Region ].Range() )
         if Diff != None:
            raise IOError, 'verification failure: ' + Diff

   def Write_Verify( self, Image, Regions = Programming_Regions ):
      "write and verify"
      self.Log( "Write_Verify( ... )" )
      for Region in Regions:
         self.Write( Image, [ Region ] )
         Readback = self.Read( [ Region ], List = Image, Ignore = 0 )
         Diff = Image.Compare(
            Readback, 'file', 'target',
            Range = self.Type.Region[ Region ].Range() )
         if Diff != None:
            raise IOError, 'verification failure: ' + Diff

   def Erase( self ):
      "erase target"
      self.Log( "Erase()" )
      if self.Type.Algorithm == Algorithm_PIC16:
         self.Bus.Send_Expect( '000ex' )
      elif self.Type.Algorithm == Algorithm_PIC16A:
         self.Bus.Send_Expect( '001ex' )
      elif self.Type.Algorithm == Algorithm_PIC16B:
         self.Bus.Send_Expect( '002ex' )
      elif self.Type.Algorithm == Algorithm_PIC16C:
         self.Bus.Send_Expect( '00ex' )
         self.Bus.Send_Expect( '001ex' )
      elif self.Type.Algorithm == Algorithm_PIC18:
         self.Bus.Send_Expect( '003ex' )
      else:
         raise IOError, \
            'internal error: unknown Algorithm ' + \
            str( PIC.Algorithm )


#############################################################################
#
# command interface
#
#############################################################################

Speed_Auto = 100
Speed_Fast = 101
Speed_Slow = 102

class Wisp_Line:
   "Wisp command line interpreter"

   def __init__( self, Line = None, Console = Dos_Console() ):
      import os
      self.Bus_Target = None
      self.Default_Baudrate = 19200
      self.Active_Baudrate = 19200
      self.Target = None
      self.ID = 0
      self.Port = 0
      if os.name == 'nt':
         self.Port = 'COM1'
      self.Image = Hex_Image()
      self.Hex = 0
      self.Lazy = 0
      self.Log = 0
      self.Patch_List = {}
      self.Protection = None
      self.Selection = [ Region_Code, Region_Data, Region_Fuses ]
      self.Verbose = 0
      self.Verify = 0
      self.Fuses = 1
      self.Fuses_Value = None
      if Line <> None:
         self.Interpret( Line )
      self.DTR = None
      self.RTS = None
      self.Console = Console
      self.Speed = Speed_Auto
      self.Last_Clock = None

   def CMD_BEEP( self ):
      import time
      self.Console.Beep()
      time.sleep( 0.2 ) 

   def CMD_BURN( self, Nr = None ):
      if Nr == None:
         Nr = self.Get_Arg()
      raise IOError, "command not implemented"

   def CMD_CALIBRATE( self ):
      raise IOError, "command not implemented"

   def CMD_CHECK( self ):
      self.Connect_If_Needed()
      self.Identify_If_Needed()
      self.Bus_Target.Verify( self.Patched_Image())

   def CMD_CLOSE( self ):
      self.Close_Bus()

   def CMD_CLEAR( self ):
      self.Image.Clear()

   def CMD_DTR( self, Value = None ):
      if Value == None:
         Value = self.Get_Arg()
      self.DTR = self.Is_Switch( Value, "DTR" )
      self.Connect_Port_If_Needed()      
      self.Bus_Target.Bus.Set_DTR( self.DTR )

   def CMD_DUMP( self ):
      self.Print( self.Patched_Image().__str__())

   def CMD_ERASE( self ):
      self.Connect_If_Needed()
      self.Identify_If_Needed()
      self.Bus_Target.Erase()

   def CMD_FLUSH( self ):
      pass

   def CMD_FUSES( self, Value = None ):
      if Value == None:
         Value = self.Get_Arg()
      if Value == 'IGNORE':
         self.Fuses = 0
         self.Fuses_Value = None
      elif Value == 'FILE':
         self.Fuses = 1
         self.Fuses_Value = None
      else:
         self.Fuses_Value = self.Hex_Value( Value )

   def CMD_GET( self ):
      self.Connect_If_Needed()
      self.Identify_If_Needed()
      self.Image = self.Bus_Target.Read()

   def CMD_GO( self ):
      self.CMD_LOAD()
      self.Connect_If_Needed()
      self.Identify_If_Needed()
      self.Bus_Target.Erase()
      self.Bus_Target.Write_Verify( self.Patched_Image())
      self.Bus_Target.Run()

   def CMD_HEX( self ):
      self.Hex = 1

   def CMD_ID( self, Value = None ):
      if Value == None:
         Value = self.Get_Arg()
      self.ID = self.Hex_Value( Value )

   def CMD_LAZY( self ):
      self.LAZY = 1

   def CMD_LOAD( self, File = None ):
      if File == None:
         File = self.Get_Arg( Uppercase = 0 )
      self.Image.Read( File )

   def CMD_LOG( self ):
      self.LOG = 1

   def CMD_PASS( self, Mode = None ):
      if Mode == None:
         Mode = self.Get_Arg()
      Message = self.Passthrough_Mode( Mode, "PASS" )
      self.Connect_If_Needed()
      self.Bus_Target.Bus.Send_Slowly( Message )
      self.Bus_Target.Bus.Send_Slowly( 'p' )
      time.sleep( 0.100 ) # wait for sending to complete

   def CMD_PATCH( self, Patch = None ):
      if Patch == None:
         Patch = self.Get_Arg()
      if Patch == 'OFF':
         self.Patch_List = {}
      if Patch == 'CLEAR':
         self.Patch_List = {}
      else:
         Split = Patch.split( ':' )
         if len( Split ) <> 2:
            IOError, "PATCH argument format error"
         self.Patch_List[ self.Hex_Value( Split[ 0 ])] = \
            self.Hex_Value( Split[ 1 ])

   def CMD_PAUSE( self, Message = None ):
      if Message == None:
         Message = self.Get_Arg( Uppercase = 0 )
      self.Print( Message )
      # <return>

   def CMD_PORT( self, Name = None ):
      self.Close_Bus()
      if Name == None:
         Name = self.Get_Arg()
      if (Name + '   ')[ 0 : 3 ] == 'COM':
         if (int((Name + '   ')[ 3 : ]) > 9):
            self.Port = '\\\\.\\' + Name
         else:
            self.Port = Name
      else:
         # added direct support for device names
         try:
            N = self.Int_Value( Name )
            if N > 32:
               self.Active_Baudrate = N
            else:
               self.Port = N
         except:
            self.Port = Name # if given as a name use it directly


   def CMD_PROTECT( self, String = None ):
      if String == None:
         String = self.Get_Arg()
      if String == 'ON':
         self.Protection = 1
      elif String == 'OFF':
         self.Protection = 0
      elif String == 'FILE':
         self.Protection = None
      else:
         raise IOError, "invalid PROTECT argument '%s'" % String

   def CMD_PUT( self ):
      self.Connect_If_Needed()
      self.Identify_If_Needed()
      self.Bus_Target.Write( self.Patched_Image(), self.Selection )

   def CMD_READ( self, File = None ):
      if File == None:
         File = self.Get_Arg()
      self.CMD_GET()
      self.CMD_SAVE( File )

   def CMD_RTS( self, Value = None ):
      if Value == None:
         Value = self.Get_Arg()
      self.RTS = self.Is_Switch( Value, "RTS" )
      self.Connect_Port_If_Needed()      
      self.Bus_Target.Bus.Set_RTS( self.RTS )

   def CMD_RUN( self ):
      self.Connect_If_Needed()
      self.Bus_Target.Run()

   def CMD_SAVE( self, File = None ):
      if File == None:
         File = self.Get_Arg( Uppercase = 0 )
      self.Patched_Image().Write( File )

   def CMD_SELECT( self, Selection = None ):
      if Selection == None:
         Selection = self.Get_Arg()
      Add = 1
      for Char in Selection:
         if Char == '+':
            Add = 1
         elif Char == '-':
            Add = 0
         elif Char == 'C':
            self.Select( Add, [ Region_Code ])
         elif Char == 'D':
            self.Select( Add, [ Region_Data ])
         elif Char == 'F':
            self.Select( Add, [ Region_Fuses ])
         else:
            raise IOError, "invalid SELECT argument char '%s'" % Char
      #print self.Selection

   def CMD_SPEED( self, Selection = None ):
      if Selection == None:
         Selection = self.Get_Arg()
      if Selection == 'FAST':
         self.Speed = Speed_Fast
      elif Selection == 'SLOW':
         self.Speed = Speed_Slow
      elif Selection == 'AUTO':
         self.Speed = Speed_Auto
      else:
         raise IOError, "invalid SPEED argument '%s'" % Selection
      self.Update_Speed()
         
   def CMD_TALK( self ):
      self.Connect_If_Needed()
      self.Emulate_TTY()

   def CMD_TARGET( self, Target = None ):
      if Target == None:
         Target = self.Get_Arg()
      if Target == 'AUTO':
         self.Target = None
      else:
         self.Target = PIC_Type_From_Name( Target, None )
         if self.Target == None:
            raise IOError, "invalid TARGET argument '%s'" % Target
      self.Image.Stride = self.Target.Stride

   def CMD_TERM( self, Baudrate = None ):
      if Baudrate == None:
         Baudrate = self.Get_Arg()
      Rate = self.Baudrate_Value( Baudrate, "TERM" )
      self.Connect_Port( Rate )
      self.Emulate_TTY( Text = "TERMinal at raw port, %d baud" % Rate )
      self.Close_Bus()

   def CMD_TEST( self ):
      raise IOError, "command not implemented"

   def CMD_TIME( self ):
      import time
      Message = ( time.asctime() )
      Now = time.clock()
      if self.Last_Clock <> None:
         Message = Message + ( " [%3.3f]" % ( Now - self.Last_Clock ))
      self.Last_Clock = Now
      self.Print( Message )

   def CMD_TTY( self, Mode = None, Baudrate = None ):
      import time
      if Mode == None:
         Mode = self.Get_Arg()
      if Baudrate <> None:
         Message = self.Passthrough_Mode( Mode, "PASS" )
      else:
         try:
            Message = self.Passthrough_Mode( Mode, "PASS" )
         except:
            Message = '0000'
            Baudrate = Mode
      if Baudrate == None:
         Baudrate = self.Get_Arg()
      Rate = self.Baudrate_Value( Baudrate, "TTY" )
      self.Connect_If_Needed()
      self.Bus_Target.Bus.Send_Expect( Message )
      self.Bus_Target.Bus.Send_Slowly( 'p' )
      time.sleep( 0.100 ) # wait for sending to complete
      self.Bus_Target.Bus.Port.setBaudrate( Rate )
      self.Emulate_TTY( Text = "TTY to target, %d baud" % Rate )

   def CMD_VCC( self ):
      raise IOError, "command not implemented"

   def CMD_VERBOSE( self ):
      self.Verbose = 1

   def CMD_VERIFY( self ):
      self.Verify = 1

   def CMD_WAIT( self, Time = None ):
      import time
      if Time == None:
         Time = self.Get_Arg()
      time.sleep( 0.001 * int( Time ))

   def CMD_WRITE( self, File = None ):
      if File == None:
         File = self.Get_Arg()
      self.CMD_LOAD( File )
      self.CMD_PUT()

   def CMD_HELP( self ):
      Text = (
         'BEEP      : beep at end               ',
         'BURN n    : set device id             ',
         'CALIBRATE : calibrate VCC             ',
         'CLEAR     : clear image               ',
         'CLOSE     : close serial connection   ',
         'CHECK     : buffer against target     ',
         'DTR x     : set DTR line              ',
         'DUMP      : dump image                ',
         'ERASE     : erase targets             ',
         'FLUSH     : flush logging             ',
         'FUSES x   : x=IGNORE, FILE or value   ',
         'GET       : target to buffer          ',
         'GO f      : erase, write f, check, run',
         'HELP      : show this text            ',
         'HEX       : hex values (TALK,TERM,TTY)',
         'ID n      : use device with ID n      ',
         'LAZY      : use lazy programming      ',
         'LOAD f    : file f to buffer          ',
         'LOG f     : log to file f             ',
         'PASS m    : enable passthrough        ',
         '            m=B6T, B6I, AUXT, AUXI    ',
         'PATCH OFF : clear patch list          ',
         'PATCH a:v : patch address a, value v  ',
         'PAUSE m   : print m, wait for return  ',
         'PORT P    : use serial port p         ',
         'PORT b    : use active baudrate b     ',
         'PROTECT x : x=ON, OFF or FILE         ',
         'PUT       : buffer to target          ',
         'READ f    : get, save f               ',
         'RTS x     : set RTS line              ',
         'RUN       : put target in run mode    ',
         'SAVE f    : buffer to file            ',
         'SELECT x  : x[i] from +-CDFA          ',
         'TALK      : talk to wisp              ',
         'TARGET x  : specify target chip       ',
         'TERM b    : TTY @ line, baudrate b    ',
         'TEST      : test programmability      ',
         'TIME      : show current time         ',
         'TTY b     : TTY to target, baudrate b ',
         'TTY [m] b : idem, for m see PASS      ',
         'VCC l h   : set verify voltages       ',
         'VERBOSE   : enable screen logging     ',
         'VERIFY f  : load f, check             ',
         'WAIT n    : wait (at least) N ms      ',
         'WRITE f   : load f, put               ',
         '                                      ')
      for Index in range( 1, len( Text ) / 2 ):
         self.Print(
            Text[ Index ] +
            ' ' +
            Text[ Index + len( Text ) / 2 ] )

   def Error( self, String ):
      raise IOError, String

   def Select( self, Add, Selection ):
      for Item in Selection:
         if Item in self.Selection:
            if not Add:
               self.Selection.remove( Item )
         else:
            if Add:
               self.Selection.append( Item )

   def Hex_Value( self, String ):
      return int( String, 16 )

   def Int_Value( self, String ):
      return int( String )

   def Close_Bus( self ):
      if self.Bus_Target != None:
         self.Bus_Target.Close()
         self.Bus_Target = None

   def Connect_Port( self, Baudrate = None ):
      if Baudrate == None:
         Baudrate = self.Default_Baudrate
      self.Close_Bus()
      self.WBus = WBus(
         Port = self.Port,
         Console = self.Console,
         ID = self.ID,
         Baudrate = Baudrate,
         DTR = self.DTR,
         RTS = self.RTS )
      self.Bus_Target = Bus_Target(
         Bus = self.WBus,
         Console = self.Console )
      if self.Verbose:
         self.Bus_Target.Debug = 1
         self.Bus_Target.Bus.Debug = 1
         
   def Connect_Port_If_Needed( self ):
      if self.Bus_Target == None:
         self.Connect_Port()

   def Connect_If_Needed( self ):
      import time
      if self.Bus_Target == None:
         self.Connect_Port()
      if not self.Bus_Target.Connected:
         self.Bus_Target.Connect()

      if self.Active_Baudrate <> self.Default_Baudrate:
         if ( self.Bus_Target.Bus.Type == 'Wisp628' ) and \
            ( float( self.Bus_Target.Bus.Version ) > 1.02 ):
            if self.Active_Baudrate == 9600:
               self.Bus_Target.Bus.Send_Expect( '0004' )
            elif self.Active_Baudrate == 19200:
               self.Bus_Target.Bus.Send_Expect( '0014' )
            elif self.Active_Baudrate == 38400:
               self.Bus_Target.Bus.Send_Expect( '0024' )
            elif self.Active_Baudrate == 57600:
               self.Bus_Target.Bus.Send_Expect( '0034' )
            elif self.Active_Baudrate == 115200:
               self.Bus_Target.Bus.Send_Expect( '0044' )
            else:
               raise self.WBus_protocol_Error, \
                  ( "bus rate %d not supported" % self.Active_Baudrate )
            self.Bus_Target.Bus.Send_Char( 'p' )
            time.sleep( 0.05 )
            self.Bus_Target.Bus.Port.setBaudrate( self.Active_Baudrate )
            self.Bus_Target.Bus.Clear()
            self.Bus_Target.Bus.Send_Expect( 'tv' )
         else:
            self.Print( 'bus speed setting not supported' )
            
      self.Update_Speed()
      Speed = ''
      if self.Bus_Target.Bus.Fast:
         Speed = '(fast)'
      self.Print(
         'hardware: %s %s %s' % (
         self.Bus_Target.Bus.Type,
         self.Bus_Target.Bus.Version,
         Speed ))

   def Identify_If_Needed( self ):
      if self.Target == None:
         self.Bus_Target.Identify()
         if self.Bus_Target.Type == None:
            self.Error(
               ( 'device word %04X or %04X, can not identify this target: ' +
               'defect, unsupported, or 16x84' ) %
               ( self.Bus_Target.Device_Code_16,
               self.Bus_Target.Device_Code ))
         else:
            self.Print(
               'target: %s, revision %02X' % (
               self.Bus_Target.Type.Name,
               self.Bus_Target.Device_Revision  ))
      else:
         self.Bus_Target.Identify()
         if self.Bus_Target.Type == None:
            self.Print(
               'target can not be verified (ID reads %04X or %04X), %s assumed'
               % ( self.Bus_Target.Device_Code_16,
                  self.Bus_Target.Device_Code,
                  self.Target.Name ))
            self.Bus_Target.Type = self.Target
         else:
            if self.Bus_Target.Type.Name == self.Target.Name:
               self.Print(
                  'target confirmed: %s, revision %02X' % (
                  self.Bus_Target.Type.Name,
                  self.Bus_Target.Device_Revision  ))
            else:
               self.Error(
                  'target specified as %s but appears to be a %s'
                  % (  self.Target.Name, self.Bus_Target.Type.Name ))
      self.Image.Stride = self.Bus_Target.Type.Stride
      self.Target = self.Bus_Target.Type

   def Update_Speed( self ):
      if self.Bus_Target <> None:
         if self.Speed == Speed_Fast:
            self.Bus_Target.Bus.Fast = 1
         elif self.Speed == Speed_Slow:
            self.Bus_Target.Bus.Fast = 0
         else:
            self.Bus_Target.Bus.Fast = 0
            if self.Bus_Target.Bus.Type == 'Wisp628':
               if float( self.Bus_Target.Bus.Version ) > 0.9:
                  self.Bus_Target.Bus.Fast = 1

   def Print( self, String ):
      self.Console.Print( String )

   def Log( self, String ):
      self.Print( String )

   def Patched_Image( self ):
      Image = self.Image.Clone()
      # patch
      for Address in self.Patch_List.keys():
         Image.Set( Address, self.Patch_List[ Address ])
      # fuses
      if self.Fuses_Value != None:
         Image.Set( self.Target.Fuses_Address, self.Fuses_Value )
      if not self.Fuses:
         Image.Erase( self.Target.Fuses_Address )
      # protection
      return Image

   def Is_Switch( self, Value, Command ):
      if Value == '+':
         return 1
      if Value == '-':
         return 0
      IOError, "%s argument error (%s)" % ( Command, Value )

   def Passthrough_Mode( self, Mode, Command ):
      if Mode == 'B6T':
         return '0000'
      elif Mode == 'B6I':
         return '0001'
      elif Mode == 'AUXT':
         return '0002'
      elif Mode == 'AUXI':
         return '0003'
      raise IOError, "%s mode argument error (%s)" % ( Command, Mode )

   def Baudrate_Value( self, Baudrate, Command ):
      try:
         return int( Baudrate )
      except:
         List = Baudrate.upper().split( "K" )
         if len( List ) == 2:
            try:
               Rate = int( List[ 1 ] )
               while ( Rate <> 0 ) and ( Rate < 100 ):
                  Rate = 10 * Rate
               return 1000 * int( List[ 0 ] ) + Rate
            except:
                pass
      raise IOError, "%s baudrate argument error (%s)" % ( Command, Baudrate )

   def Receive_Char( self ):
      if self.Bus_Target.Bus.Port.inWaiting() == 0:
         return ''
      return self.Bus_Target.Bus.Port.read( 1 )

   def Emulate_TTY( self, Text = 'TTY mode' ):
      import sys, time
      Escapes = 0
      self.Console.Set_Raw()
      self.Console.Print( '%s (ESC ESC to exit)' % Text )
      while Escapes < 2:
         Char = self.Console.Get_Key()
         if Char <> '':
            self.Bus_Target.Bus.Port.write( Char )
            if Char == chr( 27 ):
               Escapes = Escapes + 1
            else:
               Escapes = 0
         else:
            Received = self.Receive_Char()
            if Received <> '':
               self.Console.Print_String( Received )
            else:
               time.sleep( 0.010 )
      self.Console.Print( '' )
      self.Console.Print( 'TTY mode end' )
      self.Console.Print( '' )
      self.Console.Set_Cooked()

   def Get_Arg( self, Uppercase = 1 ):
      Result = ''
      while Result == '':
         if len( self.Args ) == 0:
            raise IOError, "missing argument(s) for last command"
         Result = self.Args.pop( 0 ).strip()
#      if Uppercase:
#         Result = Result.upper()
      return Result

   def Execute( self, Command ):
      if Command.startswith( 'COM' ):
         exec( 'self.CMD_PORT("' + Command + '")' )
      else:
         exec( 'self.CMD_' + Command.upper() + '()' )

   def Interpret( self, Line = None, Line_Mode = 1 ):
      import sys
      if Line == None:
         Line = sys.argv[1:]
      Line = Line[:]
      if Line_Mode:
         self.Print( 'XWisp ' + Version +', command line mode' )
      else:
         self.Print( 'XWisp ' + Version  )      
      if len( Line ) == 0:
         self.CMD_HELP()
      else:
         self.Args = Line
         while len( self.Args ) > 0:
            Command = self.Get_Arg()
            #try:
            self.Execute( Command )
            #except AttributeError:
            #   raise IOError, "unknown command '%s'" % Command
         self.Print( '%-32s' % 'OK' )


#############################################################################
#
# GUI
#
#############################################################################

class Command:
   def __init__( self, func, *args, **kw ):
      self.func = func
      self.args = args
      self.kw = kw
   def __call__( self, *args, **kw ):
      args = self.args + args
      kw.update( self.kw )
      apply( self.func, args, kw )

class Text_Window:

   def __init__( self, Parent ):
      import Tkinter
      self.Parent = Parent
      self.Root = Tkinter.Toplevel( Parent )
      self.Root.bind( '<Destroy>', Command( self.Destroy ))
      self.Text = Tkinter.Text( self.Root, height=26, width=80, font='Courier 8' )
      self.Scroll = Tkinter.Scrollbar( self.Root, command=self.Text.yview )
      self.Text.configure( yscrollcommand=self.Scroll.set )
      self.Text.pack( side=Tkinter.LEFT )
      self.Scroll.pack( side=Tkinter.RIGHT, fill=Tkinter.Y )
      self.Last_Line = None
      self.Text.focus_force()
      self.Text.bind( '<KeyPress>', Command( self.Key_Press ))
      self.Key = ''
      self.Destroyed = 0
      self.Set_Cooked()

   def Beep( self ):
      try:
         import winsound
         winsound.Beep( 440, 400 )
      except:
         self.Print( "sorry, could not beep" )

   def Destroy( self, Event ):
      #print "destroy"
      self.Destroyed = 1

   def Key_Press( self, Event ):
      #print "[%s]" % Event.keysym
      if Event.keysym == 'BackSpace':
         self.Key = chr( 8 )
      elif Event.keysym == 'Tab':
         self.Key = chr( 9 )
      elif Event.keysym == 'LineFeed':
         self.Key = chr( 10 )
      elif Event.keysym == 'Return':
         self.Key = chr( 13 )
      elif Event.keysym == 'Escape':
         self.Key = chr( 27 )
      elif Event.keysym == 'Delete':
         self.Key = chr( 127 )
      elif Event.keysym == 'Cancel':
         self.Key = chr( 27 )
      elif Event.keysym == 'space':
         self.Key = ' '
      elif len( Event.keysym ) == 1:
         self.Key = Event.keysym
      else:
         return None
      return "break"

   def Print( self, String, Progress = 0 ):
      import Tkinter
      if self.Last_Line <> None:
         self.Text.delete( self.Last_Line, Tkinter.END)
         self.Last_Line = None
      if Progress:
         self.Last_Line = self.Text.index( Tkinter.END )
      self.Text.insert( Tkinter.END, '\n' + String, 'last' )
      self.Text.see( Tkinter.END )
      self.Text.update()
      self.Parent.update()

   def Print_String( self, String ):
      import Tkinter
      self.Text.insert( Tkinter.END, String )
      self.Text.see( Tkinter.END )
      self.Text.update()
      self.Parent.update()

   def Set_Raw( self ):
      self.Text.configure( insertontime = 600 )

   def Set_Cooked( self ):
      self.Text.configure( insertontime = 0 )

   def Get_Key( self ):
      if self.Destroyed:
         print "Esc"
         return chr( 27 )
      self.Parent.update()
      Key = self.Key
      self.Key = ''
      return Key

   def Title( self, Text ):
      self.Root.title( Text)

class Wisp_Window_Interpreter( Wisp_Line ):

   def __init__( self, Parent, Console ):
      self.Parent = Parent
      Wisp_Line.__init__( self, Console = Console )

   def Run( self, Command ):
      Wisp_Line.Interpret( self, Command, Line_Mode = 0 )


class Wisp_Button:

   def __init__( self, Parent, Console, At, Text, Command ):
      import Tkinter
      self.Parent = Parent
      self.Console = Console
      self.Command = Command
      self.Text = Text
      self.Button = Tkinter.Button(
         Parent.Root,
         text = Text,
         command = self.Pressed )
      if At == '':
         self.Button.grid()
      else:
         self.Button.grid( row = At[ 0 ], column = At[ 1 ], \
            padx = 3, pady = 3,
            sticky = 'EW' )

   def Pressed( self ):
      import Tkinter
      Out = self.Console
      if self.Console == None:
         Out = Text_Window( self.Parent.Root )
         Out.Title( self.Text + '  (' + \
            reduce( lambda x,y: x + ' ' + y, self.Command ) + ')' )
      Interpreter = Wisp_Window_Interpreter( self.Parent.Root, Out )
      Interpreter.Run( self.Command )
      if self.Console == None:
         Out.Title( '(inactive)' )

class Wisp_Window:
   "Wisp GUI"

   def __init__( self ):
      self.Root = None
      self.Row = 0
      self.Column = -1
      self.Macros = { 'xwisp_defaults' : []}

   def Make_Root( self ):
      import Tkinter
      if self.Root == None:
         self.Root = Tkinter.Tk()
         self.Root.title( 'XWisp' )

   def Mainloop( self ):
      if self.Root <> None:
         self.Root.mainloop()

   def Run_File( self, File_Name ):
      File = open( File_Name, 'r' )
      for Line in File.readlines():
         if Line.endswith( '\n' ):
            Line = Line[:-1]
         # print 'interpret', Line
         self.Interpret( Line )
      File.close()

   def Run_File_If_Exists( self, File_Name ):
      try:
         File = open( File_Name, 'r' )
         File.close()
         self.Run_File( File_Name )
      except:
         return

   def Get_Arg( self, Uppercase = 1 ):
      Result = ''
      while Result == '':
         if len( self.Args ) == 0:
            return ''
         Result = self.Args.pop( 0 ).strip()
      if Uppercase:
         Result = Result.upper()
      return Result

   def Expand( self, List, N = 0 ):
      if N > 100:
         IOError, "macro expansion nesting exceeds 100"
      Result = []
      for Element in List:
         if self.Macros.has_key( Element):
            Result.extend( self.Expand( \
               self.Macros[ Element ], N + 1 ))
         else:
            if Element <> '':
               Result.append( Element )
      return Result

   def Interpret( self, Line = '' ):
      self.Args = Split_Quoted( Line )
      # print self.Args
      if len( self.Args ) > 0:
         if self.Args[ 0 ].upper() != 'MACRO':
            self.Args = self.Expand( self.Args )
      # print 'line', Line
      # print 'args', self.Args
      Command = self.Get_Arg()
      if ( Command + '#' )[ 0 ]<> '#':
         #try:
            exec( 'self.CMD_' + Command.upper() + '()' )
         #except AttributeError:
         #   raise IOError, "unknown command '%s' in line '%s'" % ( Command, Line )

   def Make_At( self, String ):
      Row = None
      Column = None
      if String <> None:
         for Char in String:
            if Char.isalpha():
               Row = ord( Char ) - ord( 'A' )
            if Char.isdigit():
               if Column == None:
                  Column = 0
               Column = 10 * Column + int( Char )
      if Row == None:
         Row = self.Row
      if Column == None:
         Column = self.Column + 1
      self.Row = Row
      self.Column = Column
      return ( Row, Column )

   def CMD_BUTTON( self ):
      Arg = self.Get_Arg()

      if Arg == 'AT':
         At = self.Get_Arg()
         Arg = self.Get_Arg()
      else:
         At = None
      At = self.Make_At( At )

      if Arg == 'TEXT':
         Text = self.Get_Arg( Uppercase = 0 ).replace( '_', ' ' )
         Arg = self.Get_Arg()
      else:
         Text = '<button>'

      if Arg == 'XWISP':
         Arg = ''
         Xwisp_Command = self.Expand( self.Macros[ 'xwisp_defaults' ][:])
         Xwisp_Command.extend( self.Expand( self.Args ))
         self.Args = []
         # print 'macros', self.Macros
         # print 'xwisp_command is', Xwisp_Command
      else:
         XWisp_Command = []

      if Arg <> '':
         raise IOError, "unknown command '%s'" % Arg

      self.Make_Root()
      Button = Wisp_Button( self, None, At, Text, Xwisp_Command )

   def CMD_LABEL( self ):
      import Tkinter
      Arg = self.Get_Arg()

      if Arg == 'AT':
         At = self.Get_Arg()
         Arg = self.Get_Arg()
      else:
         At = None
      At = self.Make_At( At )

      if Arg == 'TEXT':
         Text = self.Get_Arg( Uppercase = 0 ).replace( '_', ' ' )
         Arg = self.Get_Arg()
      else:
         Text = '<label>'

      self.Make_Root()
      Label = Tkinter.Label(
         self.Root,
         text = Text )
      if At == '':
         Label.grid()
      else:
         Label.grid( row = At[ 0 ], column = At[ 1 ], sticky = 'E' )

   def CMD_MACRO( self ):
      Name = self.Get_Arg( Uppercase = 0 )
      self.Macros[ Name ] = self.Args[:]     
      self.Args = []
      # print 'Name=', Name
      # print 'Macros=', self.Macros

   def CMD_NEXTROW( self ):
      self.Row = self.Row + 1
      self.Column = -1

   def CMD_NAME( self ):
      Name = self.Get_Arg( Uppercase = 0 )
      self.Make_Root()
      self.Root.title( Name.replace( '_', ' ' ))


#############################################################################
#
# main
#
#############################################################################

def Global_Config_File_Name():
   import sys, os
   return os.path.join( 
      os.path.abspath( os.path.dirname( sys.argv[ 0 ])),
      'XWisp.ini' )
   
def Local_Config_File_Name():
   return 'XWisp.ini'   

# this is a hack to enable window message processing for wxPython
app = None
def Set_App(wxapp):
    global app
    app = wxapp

def XWisp_Main(Line=""):
   import sys, os
   if len( Line ) == 0:
       Line = sys.argv[1:]
   if len( Line ) == 0:
      Wisp_Line().Interpret()
   else:
      File = Line[ 0 ]

      Dummy, Extension = os.path.splitext( File.upper())
      print Dummy, Extension
      if Extension == '.HEX':
         Wisp = Wisp_Window()
         Wisp.Run_File_If_Exists( Global_Config_File_Name() )
         Wisp.Run_File_If_Exists( Local_Config_File_Name() )
         File.replace( ' ', '_' )
         Wisp.Interpret( 
            'button text go_' + File.replace( ' ', '_' ) + \
            ' xwisp go "' + File + '"' )
         Wisp.Mainloop()
      elif Extension == '.XWISP':
         Wisp = Wisp_Window()
         Wisp.Run_File_If_Exists( Global_Config_File_Name() )
         Wisp.Run_File_If_Exists( Local_Config_File_Name() )
         Wisp.Run_File( File )
         Wisp.Mainloop()
      else:
         Wisp_Line().Interpret(Line)
         
if __name__ == '__main__':
   XWisp_Main()
